import lightgbm as lgb
import numpy as np
from factrainer.lightgbm import LgbDataset


class TestLgbDatasetGetitem:
    def test_row(self) -> None:
        sut = LgbDataset(dataset=lgb.Dataset(data=np.array([[1, 2, 3], [4, 5, 6]])))
        expected = LgbDataset(dataset=lgb.Dataset(data=np.array([[1, 2, 3]])))
        actual = sut[0]
        assert actual == expected

    def test_rows(self) -> None:
        sut = LgbDataset(
            dataset=lgb.Dataset(data=np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
        )
        expected = LgbDataset(
            dataset=lgb.Dataset(data=np.array([[7, 8, 9], [1, 2, 3]]))
        )
        actual = sut[[2, 0]]
        assert actual == expected



from collections.abc import Generator

from factrainer.base.dataset import IndexableDataset, RowIndex, RowsAndColumns
from sklearn.model_selection._split import _BaseKFold

import lightgbm as lgb

from .equality_checker import LgbDatasetEqualityChecker
from .slicer import LgbDatasetSlicer


class LgbDataset(IndexableDataset):
    dataset: lgb.Dataset

    def get_index(
        self, k_fold: _BaseKFold
    ) -> Generator[tuple[RowIndex, RowIndex], None, None]:
        for train_index, val_index in k_fold.split(self.dataset.data):  # type: ignore
            yield train_index.tolist(), val_index.tolist()

    def __getitem__(self, index: RowsAndColumns) -> "LgbDataset":
        match index:
            case int():
                return LgbDataset(
                    dataset=LgbDatasetSlicer(self.dataset.reference).slice(
                        self.dataset, [index]
                    )
                )
            case list():
                return LgbDataset(
                    dataset=LgbDatasetSlicer(self.dataset.reference).slice(
                        self.dataset, index
                    )
                )
            case slice():
                raise NotImplementedError
            case tuple():
                raise NotImplementedError
            case _:
                raise NotImplementedError

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, LgbDataset):
            return False
        return LgbDatasetEqualityChecker().check(self.dataset, value.dataset)



from collections.abc import Sequence
from pathlib import Path

import numpy as np
import scipy
from factrainer.base.dataset import BaseDatasetSlicer, RowIndex

import lightgbm as lgb
from lightgbm.compat import (  # type: ignore
    dt_DataTable,
    pa_Table,
    pd_DataFrame,
    pd_Series,
)

from .type import (
    LgbDataType,
    LgbGroupType,
    LgbInitScoreType,
    LgbLabelType,
    LgbPositionType,
    LgbWeightType,
)


class LgbDataSlicer(BaseDatasetSlicer[LgbDataType]):
    def slice(self, data: LgbDataType, index: RowIndex) -> LgbDataType:
        match data:
            case str():
                raise NotImplementedError
            case Path():
                raise NotImplementedError
            case np.ndarray():
                return data[index]
            case pd_DataFrame():
                return data.iloc[index]  # type: ignore
            case dt_DataTable():
                raise NotImplementedError
            case scipy.sparse.spmatrix():
                raise NotImplementedError
            case Sequence():
                raise NotImplementedError
            case pa_Table():
                raise NotImplementedError
            case _:
                raise NotImplementedError


class LgbLabelSlicer(BaseDatasetSlicer[LgbLabelType]):
    def slice(self, data: LgbLabelType, index: RowIndex) -> LgbLabelType:
        match data:
            case list():
                raise NotImplementedError
            case np.ndarray():
                return data[index]
            case pd_DataFrame():
                return data.iloc[index]
            case pd_Series():
                return data.iloc[index]
            case dt_DataTable():
                raise NotImplementedError
            case scipy.sparse.spmatrix():
                raise NotImplementedError
            case Sequence():
                raise NotImplementedError
            case pa_Table():
                raise NotImplementedError
            case _:
                raise NotImplementedError


class LgbWeightSlicer(BaseDatasetSlicer[LgbWeightType]):
    def slice(self, data: LgbWeightType, index: RowIndex) -> LgbWeightType:
        raise NotImplementedError


class LgbInitScoreSlicer(BaseDatasetSlicer[LgbInitScoreType]):
    def slice(self, data: LgbInitScoreType, index: RowIndex) -> LgbInitScoreType:
        raise NotImplementedError


class LgbGroupSlicer(BaseDatasetSlicer[LgbGroupType]):
    def slice(self, data: LgbGroupType, index: RowIndex) -> LgbGroupType:
        raise NotImplementedError


class LgbPositionSlicer(BaseDatasetSlicer[LgbPositionType]):
    def slice(self, data: LgbPositionType, index: RowIndex) -> LgbPositionType:
        raise NotImplementedError


class LgbDatasetSlicer(BaseDatasetSlicer[lgb.Dataset]):
    def __init__(self, reference: lgb.Dataset | None = None) -> None:
        self.reference = reference

    def slice(self, data: lgb.Dataset, index: RowIndex) -> lgb.Dataset:
        return lgb.Dataset(
            data=LgbDataSlicer().slice(data.data, index),
            label=(
                LgbLabelSlicer().slice(data.label, index)
                if data.label is not None
                else None
            ),
            reference=self.reference,
            weight=(
                LgbWeightSlicer().slice(data.weight, index)
                if data.weight is not None
                else None
            ),
            group=(
                LgbGroupSlicer().slice(data.group, index)
                if data.group is not None
                else None
            ),
            init_score=(
                LgbInitScoreSlicer().slice(data.init_score, index)
                if data.init_score is not None
                else None
            ),
            feature_name=data.feature_name,
            categorical_feature=data.categorical_feature,
            params=data.params,
            free_raw_data=data.free_raw_data,
            position=(
                LgbPositionSlicer().slice(data.position, index)
                if data.position is not None
                else None
            ),
        )



from collections.abc import Sequence
from typing import Any

import numpy as np
from numpy import typing as npt

def r2_score(
    y_true: Sequence[Any] | npt.NDArray[np.number[Any]],
    y_pred: Sequence[Any] | npt.NDArray[np.number[Any]],
    *,
    sample_weight: Sequence[Any] | None = None,
    multioutput: str = "uniform_average",
    force_finite: bool = True,
) -> float: ...



from factrainer.base.config import BaseMlModelConfig, BasePredictConfig, BaseTrainConfig
from factrainer.base.dataset import IndexableDataset, Prediction
from factrainer.base.raw_model import RawModel
from sklearn.model_selection._split import _BaseKFold

from ..single import SingleMlModel
from ..trait import PredictorTrait, TrainerTrait
from .config import CvMlModelConfig
from .dataset import IndexedDatasets, SplittedDatasets, SplittedDatasetsIndices
from .raw_model import CvRawModels


class CvMlModel[
    T: IndexableDataset,
    U: RawModel,
    V: BaseTrainConfig,
    W: BasePredictConfig,
](TrainerTrait[T, V], PredictorTrait[T, CvRawModels[U], W]):
    def __init__(
        self,
        model_config: BaseMlModelConfig[T, U, V, W],
        k_fold: _BaseKFold,
        n_jobs_train: int | None = None,
        n_jobs_predict: int | None = None,
    ) -> None:
        self._cv_model = SingleMlModel(
            CvMlModelConfig.from_config(model_config, n_jobs_train, n_jobs_predict)
        )
        self._k_fold = k_fold

    def train(self, dataset: T) -> None:
        datasets = SplittedDatasets.create(dataset, self._k_fold)
        self._cv_indices = datasets.indices
        self._cv_model.train(datasets.train, datasets.test)

    def predict(self, dataset: T) -> Prediction:
        datasets = IndexedDatasets.create(dataset, self.cv_indices.test)
        return self._cv_model.predict(datasets)

    @property
    def raw_model(self) -> CvRawModels[U]:
        return self._cv_model.raw_model

    @property
    def cv_indices(self) -> SplittedDatasetsIndices:
        return self._cv_indices

    @property
    def n_jobs_train(self) -> int | None:
        return self._cv_model.model_config.n_jobs_train  # type: ignore

    @n_jobs_train.setter
    def n_jobs_train(self, n_jobs: int | None) -> None:
        self._cv_model.model_config.n_jobs_train = n_jobs  # type: ignore

    @property
    def n_jobs_predict(self) -> int | None:
        return self._cv_model.model_config.n_jobs_predict  # type: ignore

    @n_jobs_predict.setter
    def n_jobs_predict(self, n_jobs: int | None) -> None:
        self._cv_model.model_config.n_jobs_predict = n_jobs  # type: ignore

    @property
    def train_config(self) -> V:
        return self._cv_model.train_config

    @train_config.setter
    def train_config(self, config: V) -> None:
        self._cv_model.train_config = config

    @property
    def pred_config(self) -> W | None:
        return self._cv_model.pred_config

    @pred_config.setter
    def pred_config(self, config: W | None) -> None:
        self._cv_model.pred_config = config



from abc import ABC, abstractmethod

from factrainer.base.config import BasePredictConfig, BaseTrainConfig
from factrainer.base.dataset import BaseDataset, Prediction
from factrainer.base.raw_model import RawModel


class PredictorTrait[T: BaseDataset, U: RawModel, W: BasePredictConfig](ABC):
    @abstractmethod
    def predict(self, dataset: T) -> Prediction:
        raise NotImplementedError

    @property
    @abstractmethod
    def raw_model(self) -> U:
        raise NotImplementedError

    @property
    @abstractmethod
    def pred_config(self) -> W | None:
        raise NotImplementedError

    @pred_config.setter
    @abstractmethod
    def pred_config(self, config: W | None) -> None:
        raise NotImplementedError


class TrainerTrait[T: BaseDataset, V: BaseTrainConfig](ABC):
    @abstractmethod
    def train(self, dataset: T) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def train_config(self) -> V:
        raise NotImplementedError

    @train_config.setter
    @abstractmethod
    def train_config(self, config: V) -> None:
        raise NotImplementedError


class ValidatableTrainerTrait[T: BaseDataset, V: BaseTrainConfig](ABC):
    @abstractmethod
    def train(self, train_dataset: T, val_dataset: T | None = None) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def train_config(self) -> V:
        raise NotImplementedError

    @train_config.setter
    @abstractmethod
    def train_config(self, config: V) -> None:
        raise NotImplementedError



import lightgbm as lgb
from factrainer.core import CvMlModel, SingleMlModel
from factrainer.lightgbm import LgbDataset, LgbModelConfig, LgbTrainConfig
from sklearn.metrics import r2_score
from sklearn.model_selection import KFold, train_test_split
from sklearn.utils._bunch import Bunch


def test_cv_model(_california_housing_data: Bunch) -> None:
    dataset = LgbDataset(
        dataset=lgb.Dataset(
            _california_housing_data.data, label=_california_housing_data.target
        )
    )
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "regression"},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvMlModel(config, k_fold)
    model.train(dataset)
    y_pred = model.predict(dataset)
    metric = r2_score(_california_housing_data.target, y_pred)

    assert (metric > 0.8) and (metric < 0.85)


def test_cv_model_parallel_pred(_california_housing_data: Bunch) -> None:
    dataset = LgbDataset(
        dataset=lgb.Dataset(
            _california_housing_data.data, label=_california_housing_data.target
        )
    )
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "regression"},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvMlModel(config, k_fold, n_jobs_train=4, n_jobs_predict=4)
    model.train(dataset)
    y_pred = model.predict(dataset)
    metric = r2_score(_california_housing_data.target, y_pred)

    assert (metric > 0.8) and (metric < 0.85)


def test_single_model(_california_housing_data: Bunch) -> None:
    train_X, test_X, train_y, test_y = train_test_split(
        _california_housing_data.data,
        _california_housing_data.target,
        test_size=0.2,
        random_state=1,
    )
    train_dataset = LgbDataset(dataset=lgb.Dataset(train_X, train_y))
    val_dataset = LgbDataset(dataset=lgb.Dataset(test_X, test_y))
    test_dataset = LgbDataset(dataset=lgb.Dataset(test_X, test_y))
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "regression"},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
    )
    model = SingleMlModel(config)

    model.train(train_dataset, val_dataset)
    y_pred = model.predict(test_dataset)
    metric = r2_score(test_y, y_pred)

    assert (metric > 0.8) and (metric < 0.85)



from abc import ABCMeta
from collections.abc import Generator

import numpy as np
from numpy import typing as npt

class _MetadataRequester: ...
class BaseCrossValidator(_MetadataRequester, metaclass=ABCMeta): ...

class _BaseKFold(BaseCrossValidator, metaclass=ABCMeta):
    def split(
        self,
        X: npt.ArrayLike,
        y: npt.ArrayLike | None = None,
        groups: npt.ArrayLike | None = None,
    ) -> Generator[tuple[npt.NDArray[np.int_], npt.NDArray[np.int_]]]: ...
